import java.rmi.*;
import java.rmi.server.*;
import java.rmi.RMISecurityManager;
import java.util.*;
import java.io.*;
import java.lang.reflect.*;

import NodeInfo;
import TaskServer;
import Consumer;

/**
 * ConsumerImpl.java
 */

public class ConsumerImpl extends UnicastRemoteObject implements Consumer,CXClassLoader {

    private NodeInfo myNode;
    private NodeInfo myTaskServer;   
    private TaskServer myTaskServerReference;
    private CXRMIClassLoader myClassLoader;
    private long startTime=0;
    

    /* ****************** TASK Specific variables ******************** */ 
    private String taskClassPath=new String(".");
    private String taskClassName=new String("FibTaskSpawn");
    private boolean isLocal=false;
    private boolean isProducerAllowed=true;
    private boolean isServerAllowed=false;
    private long taskRange=3;

    /******************** constructors **************************/

    public ConsumerImpl() throws RemoteException {
      super();
    } 

    public ConsumerImpl(String nodeInfoFilename,String topologyFilename) 
        throws RemoteException {

        super();

        //get your node info and create new task
        myNode=new NodeInfo(readNodeInfo(nodeInfoFilename));   

        //connect with the rest of the production network 
        myTaskServer=new NodeInfo(readNodeInfo(topologyFilename));

        //init your class loader 
        myClassLoader=new CXRMIClassLoader();
        myClassLoader.setClassesPath(taskClassPath);
		
        Class classRef=null;
        Constructor ct=null;
        Task myTask=null;

	  try {
            //custom constructor, etc
            classRef=myClassLoader.loadClass(taskClassName);
        
            /* **** TASK Specific constructor ******* */
            myTask = new FibTaskSpawn(5);
            dprint(2,"ConsumerImpl: task initialization OK");
			  
	  } catch (Exception e) {
		System.err.println("ConsumerImpl: error while loading class ..."+e.getMessage());		  
            e.printStackTrace();
        }

        //after creating the task you can modify some default properties if you want
        TaskInfo ti=new TaskInfo(myTask.getTaskInfo());
        ti.taskOriginator=new String(myNode.rmiRegistry);
        ti.taskOriginatorType=CXConst.consumerNode; //client
        ti.isLocal=isLocal; //I want it to get executed on the TS
        ti.isServerAllowed=isServerAllowed;
        ti.isProducerAllowed=isProducerAllowed;
        ti.jobId=CXConst.newTask; //to indicate a new task - I could use any negative number here 
        ti.taskId=new Sequence(-1,0,1,taskRange);
        myTask.setTaskInfo(ti);
        myTask.setOutputDestination(0,null,0); //this implies final results - no parent task
                    
        //set up your RMI connections ...
        if (System.getSecurityManager() == null) {
            System.setSecurityManager(new RMISecurityManager());
        }
    
        dprint(1,"ConsumerImpl: local init OK, trying to bind ...");
        String name = myNode.rmiRegistry;
        try {           
            Naming.rebind(name, this); 
            dprint(1,"ConsumerImpl bound at "+name);
        } catch (Exception e) {
            System.err.println("ConsumerImpl exception: " + 
			       e.getMessage());
            e.printStackTrace();
        }

        //now get hold of your task server and submit the task
        try {
          myTaskServerReference=getTaskServer(myTaskServer.rmiRegistry);
          dprint(2,"ConsumerImpl: ready to sign in ...");       
          myTaskServerReference.signIn(myNode);
          startTime=System.currentTimeMillis(); 
          dprint(2,"ConsumerImpl: ready to register task ..."); 
          myTaskServerReference.registerTask(taskClassName,myNode.rmiRegistry);
          dprint(2,"ConsumerImpl: ready to put task ..."); 
          myTaskServerReference.putTask(myTask.getTaskHeader(),myNode);
        } catch (Exception e) {
          System.err.println("Error while putting task to server:"+e.getMessage()); 
          e.printStackTrace();
        }
        dprint(2,"ConsumerImpl: put task OK waiting for results ...");        

    }

    /*************** consumer interface ***********************/

   /* **** TASK Specific result handling ******* */
   public boolean putResults(Result[] results, NodeInfo node)
    {
      //handle results here ...
      if (results[0]!=null)
         { 
           // note:
           // To access results at position [i] use Array.get(results[0].results,i);
           // When the returned results is a single element (not an array), i==0.
           // The Jod Id is given by the results[0].jobId parameter
           System.out.println("Fib results for job["+results[0].jobId+"] = "+Array.get(results[0].results,0));
           long timeElapsed=System.currentTimeMillis()-startTime;
           System.out.println("Estimated execution time:"+timeElapsed);

        }
      else
        {
           System.out.println("No results were returned ... ");
        }  

      return true;
    }



    public byte[] getClassBytes(String className)
	 throws RemoteException, ClassNotFoundException
	  {
		dprint(3,"ConsumerClassLoader: Class ["+className+"] was requested");
		return myClassLoader.getBytes(className);
	  }


    /*************** utilities ******************/

    private TaskServer getTaskServer(String rmiRegistry)
     {
       if (System.getSecurityManager() == null) { System.setSecurityManager(new RMISecurityManager()); } try { TaskServer tServer = (TaskServer) Naming.lookup(rmiRegistry); return tServer; } catch (Exception e) { System.err.println("getServer exception: " + e.getMessage()); return null; } 
     }

    private NodeInfo readNodeInfo(String filename) 
     {
        try {
          FileInputStream in = new FileInputStream(filename);
          ObjectInputStream s = new ObjectInputStream(in);
          NodeInfo node = (NodeInfo) s.readObject();
          return node;
        } catch (Exception e) {
          System.err.println("ConsumerImpl: Error while reading node info from file:"+filename);
          e.getMessage();
          e.printStackTrace();
          return null;  
        }
     }

    void dprint(int importance,String message)
     {
       if (importance<3)
         System.out.println(message);
     }
  
    /************* main method ******************/

    public static void main(String args[]) {
       
        try {
   	    new ConsumerImpl(args[0],args[1]); 
        } catch (Exception e) {
          System.err.println("Couldn't initiate ConsumerImpl"); 
          e.printStackTrace(); 
        }
    }

} // ConsumerImpl

